# Image-based Many-language Programming Language Identification

This is the replication package for the *Image-based Many-language Programming
Language Identification* paper. The model presented in the paper is able to
recognize programming languages starting from images rendered from source code.


## Requirements

The following software dependencies are required as prerequisites:

* [Python 3.6.8](https://www.python.org/)
* [GitHub Linguist](https://github.com/github/linguist)

The, you should install additional dependencies using pip as follows:

```
$ pip install -r requirements.txt
```

## Dataset

The starting
[dataset](https://annex.softwareheritage.org/public/dataset/content-samples/2019-10-08-github-top1k/)
contains all files extracted from the top 1000 most popular GitHub
repositories, as determined by repository "stars". It is composed of two files:

* a `.tar` file that contains all the files, identified by a SHA1 cryptographic
  checksum. It needs to be decompressed in the `dataset` folder.

* a `.gz` file that contains the information about the extension related to
  every single file present in the .tar. It needs to be placed in the
  `files_list` folder.


### Preprocessing

Our goal is to generate images representing source codes starting from the
files in the dataset. These images will be used to train the classifiers. To do
so, we need to preprocess the files in the dataset.

#### Deduplicate, find frequencies

The `preprocess.py` file contains the logic for:

* finding **duplicates** in the dataset, i.e., files that share the same SHA1
  digest
* find the number of files and the **relative frequency** for every extension
* select **non-rare extensions**, i.e. extensions that have a frequency higher
  than 10<sup>-5</sup>

To run the Python script provide the `.txt` file containing the initial list of
files (previously extracted in the `files_lists` folder, as follows:

```
$ python src/preprocess.py files_list/filenames-utf8.txt
```

#### Ground truth

To establish our *ground truth* we use the Linguist by GitHub. Since GitHub
doesn't provide a Python library for it, we created a Python command-line
wrapper to invoke Linguist.

To run the script use the following command. It is parallelized to use all
available CPUs. All the files will be put in the `linguist_dataset` folder.

```
$ python src/linguist_preprocess.py
```

#### Generate source code bundles

To generate the data that will be used to render the images we collect all the
files available for each language, looking in each of the extensions related,
and select a sub-sample of at most 1000 files.  Then we create for each
language a single text file containing of all the source code built by
appending the text of every single file.  Bundles that have too few lines are
discarded. The resulting files will be placed in the `bundles` folder.

```
$ python src/build_dataset.py
```

## Image rendering

The images used to train the model are rendered from the bundle files in the
previous step. The script takes 32 lines from the file and renders them into a
simple white text on black background image.

The first part of the bundle file is used for the **train** and the
**validation** set allowing text overlapping. The last part is used to render
images for the **test set**.

```
$ python src/rendering.py
```

### Scramble

To understand to what extent the models where able to recognize the features
inside the images we introduces bias into the images by scrambling some
specific character classes.

Scrambling is applied only to the test set.

```
$ python src/rendering_scramble.py
```

## Model

We tested three different generic image-classification models (ResNet34,
MobileNetv2 and AlexNet), pre-trained on ImageNet, and we applied Transfer
Learning to adapt them to the task of PLI.

After rendered the images, the performance measures of the models can be
obtained by running the following [Jupyter](https://jupyter.org/) notebooks:

* `resnet34.ipynb`
* `mobilenet.ipynb`
* `alexnet.ipynb`
